import enum
from math import ceil
import os
from pathlib import Path
from typing import Union, BinaryIO, Any, Dict, cast, TYPE_CHECKING
import warnings

import numpy as np

import _libjpeg


if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset


class Version(enum.IntEnum):
    v1 = 1
    v2 = 2


COLOURSPACE = {
    "MONOCHROME1": 0,
    "MONOCHROME2": 0,
    "RGB": 1,
    "YBR_FULL": 0,
    "YBR_FULL_422": 0,
}


LIBJPEG_ERROR_CODES = {
    -1024: "A parameter for a function was out of range",
    -1025: "Stream run out of data",
    -1026: "A code block run out of data",
    -1027: "Tried to perform an unputc or or an unget on an empty stream",
    -1028: "Some parameter run out of range",
    -1029: "The requested operation does not apply",
    -1030: "Tried to create an already existing object",
    -1031: "Tried to access a non-existing object",
    -1032: "A non-optional parameter was left out",
    -1033: "Forgot to delay a 0xFF",
    -1034: ("Internal error: the requested operation is not available"),
    -1035: (
        "Internal error: an item computed on a former pass does not "
        "coincide with the same item on a later pass"
    ),
    -1036: "The stream passed in is no valid jpeg stream",
    -1037: (
        "A unique marker turned up more than once. The input stream is "
        "most likely corrupt"
    ),
    -1038: "A misplaced marker segment was found",
    -1040: (
        "The specified parameters are valid, but are not supported by "
        "the selected profile. Either use a higher profile, or use "
        "simpler parameters (encoder only). "
    ),
    -1041: (
        "Internal error: the worker thread that was currently active had "
        "to terminate unexpectedly"
    ),
    -1042: (
        "The encoder tried to emit a symbol for which no Huffman code "
        "was defined. This happens if the standard Huffman table is used "
        "for an alphabet for which it was not defined. The reaction "
        "to this exception should be to create a custom huffman table "
        "instead"
    ),
    -2046: "Failed to construct the JPEG object",
}


def decode(
    stream: Union[str, os.PathLike, bytes, bytearray, BinaryIO],
    colour_transform: int = 0,
    reshape: bool = True,
) -> np.ndarray:
    """Return the decoded JPEG data from `arr` as a :class:`numpy.ndarray`.

    .. versionchanged:: 1.2

        `stream` can now also be :class:`str` or :class:`pathlib.Path`

    Parameters
    ----------
    stream : str, pathlib.Path, bytes or file-like
        The path to the JPEG file or a Python object containing the
        encoded JPEG data. If using a file-like then the object must have
        ``tell()``, ``seek()`` and ``read()`` methods.
    colour_transform : int, optional
        The colour transform used, one of:
        | ``0`` : No transform applied (default)
        | ``1`` : RGB to YCbCr
        | ``2`` : JPEG-LS pseudo RCT or RCT
        | ``3`` : Freeform
    reshape : bool, optional
        Reshape and re-view the output array so it matches the image data
        (default), otherwise return a 1D array of ``np.uint8``.

    Returns
    -------
    numpy.ndarray
        An array of containing the decoded image data.

    Raises
    ------
    RuntimeError
        If the decoding failed.
    """
    if isinstance(stream, (str, Path)):
        with open(stream, "rb") as f:
            buffer = f.read()
    elif isinstance(stream, (bytes, bytearray)):
        buffer = stream
    else:
        # BinaryIO
        stream = cast(BinaryIO, stream)
        required_methods = ["read", "tell", "seek"]
        if not all([hasattr(stream, meth) for meth in required_methods]):
            raise TypeError(
                f"Invalid type '{type(stream).__name__}' - must be the path "
                "to a JPEG file, a buffer containing the JPEG data or an open "
                "JPEG file-like"
            )
        buffer = stream.read()

    status, out, params = _libjpeg.decode(buffer, colour_transform, as_array=True)
    status = status.decode("utf-8")
    code, msg = status.split("::::")
    code = int(code)

    if code == 0 and reshape:
        bpp = ceil(params["precision"] / 8)
        if bpp == 2:
            out = out.view("uint16")

        shape = [params["rows"], params["columns"]]
        if params["nr_components"] > 1:
            shape.append(params["nr_components"])

        return cast(np.ndarray, out.reshape(*shape))

    if code == 0 and not reshape:
        return cast(np.ndarray, out)

    if code in LIBJPEG_ERROR_CODES:
        raise RuntimeError(
            f"libjpeg error code '{code}' returned from Decode(): "
            f"{LIBJPEG_ERROR_CODES[code]} - {msg}"
        )

    raise RuntimeError(f"Unknown error code '{code}' returned from Decode(): {msg}")


def decode_pixel_data(
    src: bytes,
    ds: Union["Dataset", Dict[str, Any], None] = None,
    version: int = Version.v1,
    **kwargs: Any,
) -> Union[np.ndarray, bytearray]:
    """Return the decoded JPEG data from `arr` as a :class:`numpy.ndarray`.

    Intended for use with *pydicom* ``Dataset`` objects.

    Parameters
    ----------
    src : bytes
        A Python :class:`bytes` instance containing the encoded JPEG data.
    ds : pydicom.dataset.Dataset, optional
        A :class:`~pydicom.dataset.Dataset` containing the group ``0x0028``
        elements corresponding to the *Pixel Data*. Must contain a
        (0028,0004) *Photometric Interpretation* element with the colour
        space of the pixel data, one of ``'MONOCHROME1'``, ``'MONOCHROME2'``,
        ``'RGB'``, ``'YBR_FULL'``, ``'YBR_FULL_422'``.
    version : int, optional

        * If ``1`` (default) then return the image data as an :class:`numpy.ndarray`
        * If ``2`` then return the image data as :class:`bytearray`

    Returns
    -------
    bytearray | numpy.ndarray
        The image data as either a bytearray or ndarray.

    Raises
    ------
    RuntimeError
        If the decoding failed.
    """
    ds = {} if not ds else ds

    if version == Version.v1:
        photometric_interpretation = kwargs.get("photometric_interpretation")
        pi = ds.get("PhotometricInterpretation", photometric_interpretation)
        if not pi:
            raise ValueError(
                "The (0028,0004) Photometric Interpretation element is missing "
                "from the dataset"
            )

        try:
            transform = COLOURSPACE[pi]
        except KeyError:
            warnings.warn(
                f"Unsupported (0028,0004) Photometric Interpretation '{pi}', no "
                "colour transformation will be applied"
            )
            transform = 0

        return decode(src, transform, reshape=False)

    # Version 2
    status, out, params = _libjpeg.decode(src, colourspace=0, as_array=False)
    status = status.decode("utf-8")
    code, msg = status.split("::::")
    code = int(code)

    if code == 0:
        return cast(bytearray, out)

    if code in LIBJPEG_ERROR_CODES:
        raise RuntimeError(
            f"libjpeg error code '{code}' returned from Decode(): "
            f"{LIBJPEG_ERROR_CODES[code]} - {msg}"
        )

    raise RuntimeError(f"Unknown error code '{code}' returned from Decode(): {msg}")


def get_parameters(
    stream: Union[str, os.PathLike, bytes, bytearray, BinaryIO],
) -> Dict[str, int]:
    """Return a :class:`dict` containing JPEG image parameters.

    .. versionchanged:: 1.2

        `stream` can now also be :class:`str` or :class:`pathlib.Path`

    Parameters
    ----------
    stream : str, Path, bytes
        The path to the JPEG file, a 1D array of ``np.uint8``, or a Python
        :class:`bytes` object containing the raw encoded JPEG image.

    Returns
    -------
    dict
        A :class:`dict` containing JPEG image parameters with keys including
        ``'rows'``, ``'columns'``, ``"nr_components"`` and
        ``"precision"``.

    Raises
    ------
    RuntimeError
        If reading the encoded JPEG data failed.
    """
    if isinstance(stream, (str, Path)):
        with open(stream, "rb") as f:
            buffer = f.read()
    elif isinstance(stream, (bytes, bytearray)):
        buffer = stream
    else:
        stream = cast(BinaryIO, stream)
        buffer = stream.read()

    status, params = _libjpeg.get_parameters(buffer)
    status = status.decode("utf-8")
    code, msg = status.split("::::")
    code = int(code)

    if code == 0:
        return cast(Dict[str, int], params)

    if code in LIBJPEG_ERROR_CODES:
        raise RuntimeError(
            f"libjpeg error code '{code}' returned from GetJPEGParameters(): "
            f"{LIBJPEG_ERROR_CODES[code]} - {msg}"
        )

    raise RuntimeError(
        f"Unknown error code '{code}' returned from GetJPEGParameters(): {msg}"
    )



"""Tests for the pylibjpeg pixel data handler."""

import os
import pytest

try:
    from pydicom.encaps import generate_frames

    HAS_PYDICOM = True
except ImportError:
    HAS_PYDICOM = False

from libjpeg import get_parameters
from libjpeg.data import get_indexed_datasets, JPEG_DIRECTORY
from libjpeg.utils import LIBJPEG_ERROR_CODES


DIR_10918 = os.path.join(JPEG_DIRECTORY, "10918")
DIR_14495 = os.path.join(JPEG_DIRECTORY, "14495")


REF_DCM = {
    "1.2.840.10008.1.2.4.50": [
        # filename, (rows, columns, samples/px, bits/sample)
        ("JPEGBaseline_1s_1f_u_08_08.dcm", (100, 100, 1, 8)),
        ("SC_rgb_dcmtk_+eb+cy+np.dcm", (100, 100, 3, 8)),
        ("color3d_jpeg_baseline.dcm", (480, 640, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cr.dcm", (100, 100, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cy+n1.dcm", (100, 100, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cy+s4.dcm", (100, 100, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.51": [
        ("RG2_JPLY_fixed.dcm", (2140, 1760, 1, 12)),
        ("JPEGExtended_1s_1f_u_16_12.dcm", (1024, 256, 1, 12)),
        ("JPEGExtended_3s_1f_u_08_08.dcm", (576, 768, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.57": [
        ("JPEGLossless_1s_1f_u_16_12.dcm", (1024, 1024, 1, 12)),
    ],
    "1.2.840.10008.1.2.4.70": [
        ("JPEG-LL.dcm", (1024, 256, 1, 16)),
        ("JPEGLosslessP14SV1_1s_1f_u_08_08.dcm", (768, 1024, 1, 8)),
        ("JPEGLosslessP14SV1_1s_1f_u_16_16.dcm", (535, 800, 1, 16)),
        ("MG1_JPLL.dcm", (4664, 3064, 1, 12)),
        ("RG1_JPLL.dcm", (1955, 1841, 1, 15)),
        ("RG2_JPLL.dcm", (2140, 1760, 1, 10)),
        ("SC_rgb_jpeg_gdcm.dcm", (100, 100, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.80": [
        ("emri_small_jpeg_ls_lossless.dcm", (64, 64, 1, 16)),  # should be 16
        ("MR_small_jpeg_ls_lossless.dcm", (64, 64, 1, 16)),
        ("RG1_JLSL.dcm", (1955, 1841, 1, 15)),
        ("RG2_JLSL.dcm", (2140, 1760, 1, 10)),
    ],
    "1.2.840.10008.1.2.4.81": [
        ("CT1_JLSN.dcm", (512, 512, 1, 16)),
        ("MG1_JLSN.dcm", (4664, 3064, 1, 12)),
        ("RG1_JLSN.dcm", (1955, 1841, 1, 15)),
        ("RG2_JLSN.dcm", (2140, 1760, 1, 10)),
    ],
}


def test_get_parameters_bytes():
    """Test get_parameters() using bytes."""
    with open(os.path.join(DIR_10918, "p1", "A1.JPG"), "rb") as fp:
        data = fp.read()

    params = get_parameters(data)

    info = (257, 255, 4, 8)

    assert (info[0], info[1]) == (params["rows"], params["columns"])
    assert info[2] == params["nr_components"]
    assert info[3] == params["precision"]


def test_get_parameters_binary():
    """Test get_parameters() using binaryio."""
    with open(os.path.join(DIR_10918, "p1", "A1.JPG"), "rb") as f:
        params = get_parameters(f)

    info = (257, 255, 4, 8)

    assert (info[0], info[1]) == (params["rows"], params["columns"])
    assert info[2] == params["nr_components"]
    assert info[3] == params["precision"]


def test_get_parameters_path():
    """Test get_parameters() using a path."""
    params = get_parameters(os.path.join(DIR_10918, "p1", "A1.JPG"))

    info = (257, 255, 4, 8)

    assert (info[0], info[1]) == (params["rows"], params["columns"])
    assert info[2] == params["nr_components"]
    assert info[3] == params["precision"]

@pytest.fixture
def remove_error_code():
    msg = LIBJPEG_ERROR_CODES[-1038]
    del LIBJPEG_ERROR_CODES[-1038]
    yield
    LIBJPEG_ERROR_CODES[-1038] = msg


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
def test_get_parameters_unknown_error(remove_error_code):
    """Test get_parameters() using a path."""
    msg = (
        r"Unknown error code '-1038' returned from GetJPEGParameters\(\): "
        r"unexpected EOF while parsing the image"
    )
    with pytest.raises(RuntimeError, match=msg):
        get_parameters(b"\xFF\xD8\xFF\xD8\x01\x02\x03")


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_non_conformant_raises():
    """Test that a non-conformant JPEG image raises an exception."""
    ds_list = get_indexed_datasets("1.2.840.10008.1.2.4.51")
    # Image has invalid Se value in the SOS marker segment
    item = ds_list["JPEG-lossy.dcm"]
    assert 0xC000 == item["Status"][1]
    ds = item["ds"]

    nr_frames = ds.get("NumberOfFrames", 1)
    frame = next(generate_frames(ds.PixelData, number_of_frames=nr_frames))

    msg = (
        r"libjpeg error code '-1038' returned from GetJPEGParameters\(\): A "
        r"misplaced marker segment was found - scan start must be zero "
        r"and scan stop must be 63 for the sequential operating modes"
    )
    with pytest.raises(RuntimeError, match=msg):
        get_parameters(frame)


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
class TestGetParametersDCM:
    """Tests for get_parameters() using DICOM datasets."""

    def generate_frames(self, ds):
        """Return a generator object with the dataset's pixel data frames."""
        nr_frames = ds.get("NumberOfFrames", 1)
        return generate_frames(ds.PixelData, number_of_frames=nr_frames)

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.50"])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.50")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.51"])
    def test_extended(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.51")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.57"])
    def test_lossless(self, fname, info):
        """Test get_parameters() for the lossless datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.57")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.70"])
    def test_lossless_sv1(self, fname, info):
        """Test get_parameters() for the lossless SV1 datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.70")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.80"])
    def test_jls_lossless(self, fname, info):
        """Test get_parameters() for the LS lossless datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.80")
        ds = index[fname]["ds"]
        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.81"])
    def test_jls_lossy(self, fname, info):
        """Test get_parameters() for the LS lossy datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.81")
        ds = index[fname]["ds"]
        frame = next(self.generate_frames(ds))
        params = get_parameters(frame)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]


REF_JPG = {
    "10918": {
        "p1": [
            ("A1.JPG", (257, 255, 4, 8)),
            # ('B1.JPG', (1, 1, 1, 1)),
            # ('B2.JPG', (1, 1, 1, 1)), missing DHT marker (its in B1.JPG)
        ],
        "p2": [
            ("A1.JPG", (257, 255, 4, 8)),
            # ('B1.JPG', ()),
            # ('B2.JPG', ()), missing DHT (in B1)
            ("C1.JPG", (257, 255, 4, 8)),
            ("C2.JPG", (257, 255, 4, 8)),
        ],
        "p4": [
            ("A1.JPG", (257, 255, 4, 8)),
            # ('B1.JPG', ()),
            # ('B2.JPG', ()),
            ("C1.JPG", (257, 255, 4, 8)),
            ("C2.JPG", (257, 255, 4, 8)),
            ("E1.JPG", (257, 255, 4, 12)),
            ("E2.JPG", (257, 255, 4, 12)),
        ],
        "p14": [
            ("O1.JPG", (257, 255, 4, 8)),
            ("O2.JPG", (257, 255, 4, 16)),
        ],
    },
    "14495": {
        "JLS": [
            ("T8C0E0.JLS", (256, 256, 3, 8)),
            ("T8C1E0.JLS", (256, 256, 3, 8)),
            ("T8C2E0.JLS", (256, 256, 3, 8)),
            ("T8NDE0.JLS", (128, 128, 1, 8)),
            ("T8SSE0.JLS", (256, 256, 3, 8)),
            ("T16E0.JLS", (256, 256, 1, 12)),
        ],
        "JNL": [
            ("T8C0E3.JLS", (256, 256, 3, 8)),
            ("T8C1E3.JLS", (256, 256, 3, 8)),
            ("T8C2E3.JLS", (256, 256, 3, 8)),
            ("T8NDE3.JLS", (128, 128, 1, 8)),
            ("T8SSE3.JLS", (256, 256, 3, 8)),
            ("T16E3.JLS", (256, 256, 1, 12)),
        ],
    },
    "15444": {},
}


class TestGetParametersJPG:
    """Tests for get_parameters() using JPEG compliance data."""

    @pytest.mark.parametrize("fname, info", REF_JPG["10918"]["p1"])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p1", fname), "rb") as fp:
            data = fp.read()

        params = get_parameters(data)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG["10918"]["p2"])
    def test_extended_p2(self, fname, info):
        """Test get_parameters() for the extended p2 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p2", fname), "rb") as fp:
            data = fp.read()

        params = get_parameters(data)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG["10918"]["p4"])
    def test_extended_p4(self, fname, info):
        """Test get_parameters() for the extended p4 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p4", fname), "rb") as fp:
            data = fp.read()

        params = get_parameters(data)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG["10918"]["p14"])
    def test_lossless_p14(self, fname, info):
        """Test get_parameters() for the extended p14 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p14", fname), "rb") as fp:
            data = fp.read()

        params = get_parameters(data)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG["14495"]["JLS"])
    def test_jls(self, fname, info):
        """Test get_parameters() for the JPEG-LS compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, "JLS", fname), "rb") as fp:
            data = fp.read()

        params = get_parameters(data)

        assert (info[0], info[1]) == (params["rows"], params["columns"])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]



"""Tests for pylibjpeg.decode()."""

import os
from pathlib import Path
import pytest

import numpy as np

try:
    from pydicom.encaps import generate_frames
    from pydicom.pixels.utils import reshape_pixel_array

    HAS_PYDICOM = True
except ImportError:
    HAS_PYDICOM = False

from libjpeg import decode, decode_pixel_data
from libjpeg.data import get_indexed_datasets, JPEG_DIRECTORY


DIR_10918 = JPEG_DIRECTORY / "10918"
DIR_14495 = JPEG_DIRECTORY / "14495"


REF_DCM = {
    "1.2.840.10008.1.2.4.50": [
        # filename, (rows, columns, samples/px, bits/sample)
        ("JPEGBaseline_1s_1f_u_08_08.dcm", (100, 100, 1, 8)),
        ("SC_rgb_dcmtk_+eb+cy+np.dcm", (100, 100, 3, 8)),
        ("color3d_jpeg_baseline.dcm", (480, 640, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cr.dcm", (100, 100, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cy+n1.dcm", (100, 100, 3, 8)),
        ("SC_rgb_dcmtk_+eb+cy+s4.dcm", (100, 100, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.51": [
        ("RG2_JPLY_fixed.dcm", (2140, 1760, 1, 12)),
        ("JPEGExtended_1s_1f_u_16_12.dcm", (1024, 256, 1, 12)),
        ("JPEGExtended_3s_1f_u_08_08.dcm", (576, 768, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.57": [
        ("JPEGLossless_1s_1f_u_16_12.dcm", (1024, 1024, 1, 12)),
    ],
    "1.2.840.10008.1.2.4.70": [
        ("JPEG-LL.dcm", (1024, 256, 1, 16)),
        ("JPEGLosslessP14SV1_1s_1f_u_08_08.dcm", (768, 1024, 1, 8)),
        ("JPEGLosslessP14SV1_1s_1f_u_16_16.dcm", (535, 800, 1, 16)),
        ("MG1_JPLL.dcm", (4664, 3064, 1, 12)),
        ("RG1_JPLL.dcm", (1955, 1841, 1, 15)),
        ("RG2_JPLL.dcm", (2140, 1760, 1, 10)),
        ("SC_rgb_jpeg_gdcm.dcm", (100, 100, 3, 8)),
    ],
    "1.2.840.10008.1.2.4.80": [
        ("emri_small_jpeg_ls_lossless.dcm", (64, 64, 1, 12)),
        ("MR_small_jpeg_ls_lossless.dcm", (64, 64, 1, 16)),
        ("RG1_JLSL.dcm", (1955, 1841, 1, 16)),
        ("RG2_JLSL.dcm", (2140, 1760, 1, 10)),
    ],
    "1.2.840.10008.1.2.4.81": [
        ("CT1_JLSN.dcm", (512, 512, 1, 16)),
        ("MG1_JLSN.dcm", (4664, 3064, 1, 12)),
        ("RG1_JLSN.dcm", (1955, 1841, 1, 15)),
        ("RG2_JLSN.dcm", (2140, 1760, 1, 10)),
    ],
}


# TODO: convert using straight JPG data
@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_decode_bytes():
    """Test decode using bytes."""
    index = get_indexed_datasets("1.2.840.10008.1.2.4.50")
    ds = index["JPEGBaseline_1s_1f_u_08_08.dcm"]["ds"]
    nr_frames = ds.get("NumberOfFrames", 1)
    frame = next(generate_frames(ds.PixelData, number_of_frames=nr_frames))
    arr = decode(frame)
    assert arr.flags.writeable
    assert "uint8" == arr.dtype
    assert (ds.Rows, ds.Columns) == arr.shape

    # Reference values from GDCM handler
    assert 76 == arr[5, 50]
    assert 167 == arr[15, 50]
    assert 149 == arr[25, 50]
    assert 203 == arr[35, 50]
    assert 29 == arr[45, 50]
    assert 142 == arr[55, 50]
    assert 1 == arr[65, 50]
    assert 64 == arr[75, 50]
    assert 192 == arr[85, 50]
    assert 255 == arr[95, 50]


def test_decode_binaryio():
    """Test decode using binaryio."""
    with open(DIR_10918 / "p1" / "A1.JPG", "rb") as f:
        arr = decode(f)
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert arr.shape == (257, 255, 4)


def test_decode_raises():
    """Test decode raises if invalid type"""
    msg = (
        "Invalid type 'NoneType' - must be the path to a JPEG file, a "
        "buffer containing the JPEG data or an open JPEG file-like"
    )
    with pytest.raises(TypeError, match=msg):
        decode(None)


def test_v1_invalid_photometric_raises():
    """Test decode_pixel_data raises if invalid photometric interpretation"""
    msg = (
        r"The \(0028,0004\) Photometric Interpretation element is missing "
        "from the dataset"
    )
    with pytest.raises(ValueError, match=msg):
        decode_pixel_data(DIR_10918 / "p1" / "A1.JPG", version=1)


# TODO: convert to using straight JPG data
@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_invalid_colourspace_warns():
    """Test that using an unknown colourspace gives a warning."""
    index = get_indexed_datasets("1.2.840.10008.1.2.4.50")
    ds = index["JPEGBaseline_1s_1f_u_08_08.dcm"]["ds"]
    nr_frames = ds.get("NumberOfFrames", 1)
    frame = next(generate_frames(ds.PixelData, number_of_frames=nr_frames))
    msg = r"no colour transformation will be applied"
    ds.PhotometricInterpretation = "ANY"
    with pytest.warns(UserWarning, match=msg):
        arr = decode_pixel_data(frame, ds)

    arr = reshape_pixel_array(ds, arr)

    assert arr.flags.writeable
    assert "uint8" == arr.dtype
    assert (ds.Rows, ds.Columns) == arr.shape

    # Reference values from GDCM handler
    assert 76 == arr[5, 50]
    assert 167 == arr[15, 50]
    assert 149 == arr[25, 50]
    assert 203 == arr[35, 50]
    assert 29 == arr[45, 50]
    assert 142 == arr[55, 50]
    assert 1 == arr[65, 50]
    assert 64 == arr[75, 50]
    assert 192 == arr[85, 50]
    assert 255 == arr[95, 50]


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_decode_adobe_v101():
    """Test decoding JPEG with Adobe v101 APP14 marker."""
    index = get_indexed_datasets("1.2.840.10008.1.2.4.50")
    ds = index["SC_jpeg_no_color_transform_2.dcm"]["ds"]
    frame = next(generate_frames(ds.PixelData, number_of_frames=1))
    arr = decode(frame, reshape=True)
    assert arr.shape == (256, 256, 3)


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
class TestDecodeDCM:
    """Tests for get_parameters() using DICOM datasets."""

    def generate_frames(self, ds):
        """Return a generator object with the dataset's pixel data frames."""
        nr_frames = ds.get("NumberOfFrames", 1)
        return generate_frames(ds.PixelData, number_of_frames=nr_frames)

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.50"])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.50")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.51"])
    def test_extended(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.51")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.57"])
    def test_lossless(self, fname, info):
        """Test get_parameters() for the lossless datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.57")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.70"])
    def test_lossless_sv1(self, fname, info):
        """Test get_parameters() for the lossless SV1 datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.70")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.80"])
    def test_jls_lossless(self, fname, info):
        """Test get_parameters() for the LS lossless datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.80")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    @pytest.mark.parametrize("fname, info", REF_DCM["1.2.840.10008.1.2.4.81"])
    def test_jls(self, fname, info):
        """Test get_parameters() for the LS lossy datasets."""
        # info: (rows, columns, spp, bps)
        index = get_indexed_datasets("1.2.840.10008.1.2.4.81")
        ds = index[fname]["ds"]

        frame = next(self.generate_frames(ds))
        arr = decode(frame, reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

    def test_v2_non_conformant_raises(self):
        """Test that a non-conformant JPEG image raises an exception."""
        ds_list = get_indexed_datasets("1.2.840.10008.1.2.4.51")
        # Image has invalid Se value in the SOS marker segment
        item = ds_list["JPEG-lossy.dcm"]
        assert 0xC000 == item["Status"][1]
        ds = item["ds"]
        nr_frames = ds.get("NumberOfFrames", 1)
        frame = next(generate_frames(ds.PixelData, number_of_frames=nr_frames))
        msg = (
            r"libjpeg error code '-1038' returned from Decode\(\): A "
            r"misplaced marker segment was found - scan start must be zero "
            r"and scan stop must be 63 for the sequential operating modes"
        )
        with pytest.raises(RuntimeError, match=msg):
            decode_pixel_data(frame, version=2)


REF_JPG = {
    "10918": {
        # Reference values using libjpeg's `jpeg` app + imageJ
        # Very circular but having issues with other programs interpreting
        #   it as CYMK and doing weird stuff
        "p1": [
            # ([top left pixel], [bottom right pixel])
            ("A1.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            # ('B1.JPG', (1, 1, 1, 1)),
            # ('B2.JPG', (1, 1, 1, 1)), missing DHT marker (its in B1.JPG)
        ],
        "p2": [
            ("A1.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            # ('B1.JPG', ()),
            # ('B2.JPG', ()), missing DHT (in B1)
            ("C1.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            ("C2.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
        ],
        "p4": [
            ("A1.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            # ('B1.JPG', ()),
            # ('B2.JPG', ()),
            ("C1.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            ("C2.JPG", (257, 255, 4, 8), ([138, 76, 239, 216], [155, 217, 191, 115])),
            (
                "E1.JPG",
                (257, 255, 4, 12),
                ([2119, 1183, 3907, 3487], [2502, 3402, 3041, 1872]),
            ),
            (
                "E2.JPG",
                (257, 255, 4, 12),
                ([2119, 1183, 3907, 3487], [2502, 3402, 3041, 1872]),
            ),
        ],
        "p14": [
            ("O1.JPG", (257, 255, 4, 8), ([132, 76, 245, 218], [156, 211, 191, 116])),
            (
                "O2.JPG",
                (257, 255, 4, 16),
                ([33792, 19456, 62720, 55808], [39936, 53888, 48768, 29696]),
            ),
        ],
    },
    "14495": {
        "JLS": [  # Lossless
            (
                # Reference values using GDCM
                "T8C0E0.JLS",
                (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95]),
            ),
            (
                # GDCM and CharPyLS fail on this one - both use CharLS
                # Circular reference
                "T8C1E0.JLS",
                (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95]),
            ),
            (
                # CharLS fails
                # Circular reference
                "T8C2E0.JLS",
                (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95]),
            ),
            (
                # Reference values using GDCM
                "T8NDE0.JLS",
                (128, 128, 1, 8),
                (108, 231),
            ),
            (
                # CharLS fails
                # Circular reference
                "T8SSE0.JLS",
                (256, 256, 3, 8),
                ([161, 122, 108], [101, 171, 231]),
            ),
            (
                # Reference values using GDCM
                "T16E0.JLS",
                (256, 256, 1, 12),
                (1963, 1596),
            ),
        ],
        "JNL": [  # Near lossless
            (
                # Reference values using GDCM
                "T8C0E3.JLS",
                (256, 256, 3, 8),
                ([161, 119, 105], [98, 96, 93]),
            ),
            (
                # CharLS fails
                # Circular reference
                "T8C1E3.JLS",
                (256, 256, 3, 8),
                ([161, 119, 105], [101, 100, 97]),
            ),
            (
                # CharLS fails
                # Circular reference
                "T8C2E3.JLS",
                (256, 256, 3, 8),
                ([161, 119, 105], [98, 96, 94]),
            ),
            (
                # Reference values using GDCM
                "T8NDE3.JLS",
                (128, 128, 1, 8),
                (105, 229),
            ),
            (
                # CharLS fails
                # Circular reference
                "T8SSE3.JLS",
                (256, 256, 3, 8),
                ([161, 119, 105], [102, 169, 234]),
            ),
            (
                # Reference values using GDCM
                "T16E3.JLS",
                (256, 256, 1, 12),
                (1960, 1593),
            ),
        ],
    },
    "15444": {},
}


class TestDecodeJPG:
    """Tests for get_parameters() using JPEG compliance data."""

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["10918"]["p1"])
    def test_baseline(self, fname, info, ref):
        """Test decoding the baseline compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p1", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        # Process 1 is always 8-bit
        assert arr.dtype == "uint8"

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

        buffer = decode_pixel_data(data, version=2)
        assert isinstance(buffer, bytearray)
        arr = np.frombuffer(buffer, dtype="u1")
        assert arr.flags.writeable
        if info[2] == 1:
            arr = arr.reshape((info[0], info[1]))
        else:
            arr = arr.reshape((info[0], info[1], info[2]))

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["10918"]["p2"])
    def test_extended_p2(self, fname, info, ref):
        """Test decoding the extended p2 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p2", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        # Process 2 is always 8-bit
        assert arr.dtype == "uint8"

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["10918"]["p4"])
    def test_extended_p4(self, fname, info, ref):
        """Test decoding the extended p4 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p4", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["10918"]["p14"])
    def test_lossless_p14(self, fname, info, ref):
        """Test decoding the extended p14 compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, "p14", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["14495"]["JLS"])
    def test_jls(self, fname, info, ref):
        """Test decoding the JPEG-LS lossless compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, "JLS", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

        if info[2] == 1:
            assert ref[0] == arr[0, 0].tolist()
            assert ref[1] == arr[-1, -1].tolist()
        else:
            assert ref[0] == arr[0, 0, :].tolist()
            assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG["14495"]["JNL"])
    def test_jnl(self, fname, info, ref):
        """Test decoding the JPEG-LS near lossless compliance images."""
        # info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, "JNL", fname), "rb") as fp:
            data = fp.read()

        arr = decode(data, reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == "uint8"
        if 9 <= info[3] <= 16:
            assert arr.dtype == "uint16"

        if info[2] == 1:
            assert ref[0] == arr[0, 0].tolist()
            assert ref[1] == arr[-1, -1].tolist()
        else:
            assert ref[0] == arr[0, 0, :].tolist()
            assert ref[1] == arr[-1, -1, :].tolist()

    def test_str(self):
        """Test decoding using a str path."""
        p = DIR_10918 / "p1" / "A1.JPG"
        p = os.fspath(p)
        assert isinstance(p, str)
        arr = decode(p)
        assert arr.shape == (257, 255, 4)

    def test_path(self):
        """Test decoding using a Path path."""
        p = DIR_10918 / "p1" / "A1.JPG"
        assert isinstance(p, Path)
        arr = decode(p)
        assert arr.shape == (257, 255, 4)



"""Tests for the pylibjpeg pixel data handler."""

import pytest

try:
    from pydicom.pixels import convert_color_space
    from pydicom.encaps import generate_frames

    HAS_PYDICOM = True
except ImportError:
    HAS_PYDICOM = False

from libjpeg import decode, decode_pixel_data
from libjpeg.utils import COLOURSPACE
from libjpeg.data import get_indexed_datasets


class HandlerTestBase:
    """Baseclass for handler tests."""

    uid = None

    def setup_method(self):
        self.ds = get_indexed_datasets(self.uid)

    def plot(self, arr, index=None, cmap=None):
        import matplotlib.pyplot as plt

        if index is not None:
            if cmap:
                plt.imshow(arr[index], cmap=cmap)
            else:
                plt.imshow(arr[index])
        else:
            if cmap:
                plt.imshow(arr, cmap=cmap)
            else:
                plt.imshow(arr)

        plt.show()


@pytest.fixture
def add_invalid_colour():
    COLOURSPACE["INVALID"] = -1
    yield
    del COLOURSPACE["INVALID"]


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestLibrary:
    """Tests for libjpeg itself."""

    def test_non_conformant_raises(self):
        """Test that a non-conformant JPEG image raises an exception."""
        ds_list = get_indexed_datasets("1.2.840.10008.1.2.4.51")
        # Image has invalid Se value in the SOS marker segment
        item = ds_list["JPEG-lossy.dcm"]
        assert 0xC000 == item["Status"][1]
        msg = (
            r"libjpeg error code '-1038' returned from Decode\(\): A "
            r"misplaced marker segment was found - scan start must be zero "
            r"and scan stop must be 63 for the sequential operating modes"
        )
        with pytest.raises(RuntimeError, match=msg):
            item["ds"].pixel_array

    def test_invalid_colour_transform(self, add_invalid_colour):
        """Test that an invalid colour transform raises an exception."""
        ds_list = get_indexed_datasets("1.2.840.10008.1.2.4.50")
        # Image has invalid Se value in the SOS marker segment
        ds = ds_list["color3d_jpeg_baseline.dcm"]["ds"]
        data = next(generate_frames(ds.PixelData, number_of_frames=1))
        msg = (
            r"Unknown error code '-8194' returned from Decode\(\): "
            r"Invalid colourTransform value"
        )
        with pytest.raises(RuntimeError, match=msg):
            decode(data, -1)


# ISO/IEC 10918 JPEG
@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGBaseline(HandlerTestBase):
    """Test the handler with ISO 10918 JPEG images.

    1.2.840.10008.1.2.4.50 : JPEG Baseline (Process 1)

    Process 1: Baseline sequential DCT, 8-bit sample precision.

    Image samples assumed to be unsigned in the range 0 to 255.

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2, YBR_FULL_422
    Samples Per Pixel: 1 or 3
    Planar Configuration: 0 (for YBR_FULL_422)
    Pixel Representation: 0 (unsigned)
    Bits Allocated/Stored: 8/8

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.1-1
      <part05/sect_8.2.html#table_8.2.1-1>`
    * ISO/IEC 10918-1 (ITU T.81)
    * ISO/IEC 10918-2
    * `JPEG Image Compression Systems
      <https://www.ece.ucdavis.edu/cerl/reliablejpeg/compression/>`_

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.50"

    def test_1s_1f(self):
        """Test greyscale."""
        ds = self.ds["JPEGBaseline_1s_1f_u_08_08.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # Reference values from GDCM handler
        assert 76 == arr[5, 50]
        assert 167 == arr[15, 50]
        assert 149 == arr[25, 50]
        assert 203 == arr[35, 50]
        assert 29 == arr[45, 50]
        assert 142 == arr[55, 50]
        assert 1 == arr[65, 50]
        assert 64 == arr[75, 50]
        assert 192 == arr[85, 50]
        assert 255 == arr[95, 50]

    @pytest.mark.skip("No suitable dataset")
    def test_1s_Nf(self):
        """Test greyscale with N frames."""
        ds = self.ds["JPEGBaseline_1s_2f_08_08.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 2 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (2, ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr, index=0, cmap='gray')

    def test_3s_1f_ybr_422(self):
        """Test YBR with 422 subsampling."""
        # +cy is YCbCr
        # +s2 is 422 subsampling w/ YBR_FULL_422
        ds = self.ds["SC_rgb_dcmtk_+eb+cy+np.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "YBR_FULL_422" == ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        ds.pixel_array_options(raw=True)
        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        # Reference values may be slightly different depending on handler
        # Pillow: assert ( 76,  86, 251) == tuple(arr[ 5, 50, :])
        assert (76, 85, 254) == tuple(arr[5, 50, :])
        # GDCM:   assert (166, 109, 190) == tuple(arr[15, 50, :])
        # Pillow: assert (166, 112, 185) == tuple(arr[15, 50, :])
        assert (166, 108, 190) == tuple(arr[15, 50, :])
        # Pillow: assert (150,  46,  17) == tuple(arr[25, 50, :])
        assert (150, 46, 21) == tuple(arr[25, 50, :])
        # GDCM:   assert (203,  85,  74) == tuple(arr[35, 50, :])
        # Pillow: assert (203,  95,  75) == tuple(arr[35, 50, :])
        assert (203, 86, 74) == tuple(arr[35, 50, :])
        # GDCM:   assert ( 29, 255, 108) == tuple(arr[45, 50, :])
        # Pillow: assert ( 29, 255, 109) == tuple(arr[45, 50, :])
        assert (29, 255, 107) == tuple(arr[45, 50, :])
        # Pillow: assert (142, 189, 118) == tuple(arr[55, 50, :])
        assert (142, 192, 117) == tuple(arr[55, 50, :])
        assert (0, 128, 128) == tuple(arr[65, 50, :])
        assert (64, 128, 128) == tuple(arr[75, 50, :])
        assert (192, 128, 128) == tuple(arr[85, 50, :])
        assert (255, 128, 128) == tuple(arr[95, 50, :])

    def test_3s_Nf_ybr_422(self):
        """Test 3 sample/px with N frames."""
        ds = self.ds["color3d_jpeg_baseline.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 != getattr(ds, "NumberOfFrames", 1)
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert "YBR_FULL_422" == ds.PhotometricInterpretation
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.NumberOfFrames, ds.Rows, ds.Columns, 3) == arr.shape

        # Gives a MemoryError in GitHub CI build otherwise
        arr = arr[0:10, ...]

        # GDCM: all match
        assert (41, 41, 41) == tuple(arr[3, 159, 290, :])
        assert (57, 57, 57) == tuple(arr[3, 169, 290, :])
        # Pillow: assert (71, 168, 125) == tuple(arr[3, 41, 380, :])
        assert (72, 167, 125) == tuple(arr[3, 41, 380, :])

    # Non-conformant datasets
    def test_3s_1f_rgb_non(self):
        """Test RGB."""
        # +cr is RGB - DICOM non-conformant
        ds = self.ds["SC_rgb_dcmtk_+eb+cr.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "RGB" == ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        # self.plot(arr)

        assert (255, 0, 0) == tuple(arr[5, 50, :])
        assert (255, 128, 128) == tuple(arr[15, 50, :])
        assert (0, 255, 0) == tuple(arr[25, 50, :])
        assert (128, 255, 128) == tuple(arr[35, 50, :])
        assert (0, 0, 255) == tuple(arr[45, 50, :])
        assert (128, 128, 255) == tuple(arr[55, 50, :])
        assert (0, 0, 0) == tuple(arr[65, 50, :])
        assert (64, 64, 64) == tuple(arr[75, 50, :])
        assert (192, 192, 192) == tuple(arr[85, 50, :])
        assert (255, 255, 255) == tuple(arr[95, 50, :])

    def test_3s_1f_ybr_411_non(self):
        """Test YBR with 411 subsampling."""
        # +cy is YCbCr
        # +n1 is 411 subsampling w/ YBR_FULL - DICOM non-conformant
        ds = self.ds["SC_rgb_dcmtk_+eb+cy+n1.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "YBR_FULL" == ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        # GDCM: all match
        # Pillow: assert (248,   3,   2) == tuple(arr[ 5, 50, :])
        assert (253, 1, 0) == tuple(arr[5, 50, :])
        # Pillow: assert (246, 131, 138) == tuple(arr[15, 50, :])
        assert (253, 129, 131) == tuple(arr[15, 50, :])
        # Pillow: assert (129, 252, 145) == tuple(arr[35, 50, :])
        assert (127, 255, 129) == tuple(arr[35, 50, :])
        # Pillow: assert (  2,   0, 254) == tuple(arr[45, 50, :])
        assert (0, 0, 254) == tuple(arr[45, 50, :])
        # Pillow: assert (128, 128, 250) == tuple(arr[55, 50, :])
        assert (127, 128, 255) == tuple(arr[55, 50, :])
        assert (0, 0, 0) == tuple(arr[65, 50, :])
        assert (64, 64, 64) == tuple(arr[75, 50, :])
        assert (192, 192, 192) == tuple(arr[85, 50, :])
        assert (255, 255, 255) == tuple(arr[95, 50, :])

    def test_3s_1f_ybr_444_non(self):
        """Test YBR with 444 subsampling."""
        # +cy is YCbCr
        # +s4 is 444 subsampling w/ YBR_FULL - DICOM non-conformant
        ds = self.ds["SC_rgb_dcmtk_+eb+cy+s4.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "YBR_FULL" == ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        assert (254, 0, 0) == tuple(arr[5, 50, :])
        assert (255, 127, 127) == tuple(arr[15, 50, :])
        assert (0, 255, 5) == tuple(arr[25, 50, :])
        assert (129, 255, 129) == tuple(arr[35, 50, :])
        assert (0, 0, 254) == tuple(arr[45, 50, :])
        assert (128, 127, 255) == tuple(arr[55, 50, :])
        assert (0, 0, 0) == tuple(arr[65, 50, :])
        assert (64, 64, 64) == tuple(arr[75, 50, :])
        assert (192, 192, 192) == tuple(arr[85, 50, :])
        assert (255, 255, 255) == tuple(arr[95, 50, :])


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGExtended(HandlerTestBase):
    """Test the handler with ISO 10918 JPEG images.

    1.2.840.10008.1.2.4.51 : JPEG Extended (Process 2 and 4)

    Process 2: Extended sequential DCT, Huffman coding, 8-bit sample precision
    Process 4: Extended sequential DCT, Huffman coding, 12-bit sample precision

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2
    Samples Per Pixel: 1
    Pixel Representation: 0 (unsigned)
    Bits Allocated/Stored: 8/8 or 16/12

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.1-1
      <part05/sect_8.2.html#table_8.2.1-1>`
    * ISO/IEC 10918-1 (ITU T.81)
    * ISO/IEC 10918-2
    * `JPEG Image Compression Systems
      <https://www.ece.ucdavis.edu/cerl/reliablejpeg/compression/>`_

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/12 bits allocated/stored, unsigned, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.51"

    # Process 2
    @pytest.mark.skip("No suitable dataset")
    def test_1s_1f_u_08_08(self):
        pass

    @pytest.mark.skip("No suitable dataset")
    def test_1s_Nf_u_08_08(self):
        pass

    # Process 4
    def test_1s_1f_u_16_10(self):
        """Test process 4 greyscale."""
        ds = self.ds["RG2_JPLY_fixed.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        # Input precision is 12, not 10
        assert 10 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from output of IJG's djpeg recompiled for 12-bit
        # Note differences       |              |
        # IJG: [571, 582, 666, 766, 803, 781, 726, 651, 595, 578]
        assert [571, 582, 666, 765, 803, 781, 725, 651, 595, 578] == (
            arr[845, 964:974].tolist()
        )

    def test_1s_1f_u_16_12(self):
        """Test process 4 greyscale."""
        ds = self.ds["JPEGExtended_1s_1f_u_16_12.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 12 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from output of IJG's djpeg recompiled for 12-bit
        # Note differences  |
        # IJG: [134, 143, 161, 172, 180, 193, 249, 214, 223, 247]
        assert [134, 143, 160, 172, 180, 193, 249, 214, 223, 247] == (
            arr[410:420, 136].tolist()
        )

    @pytest.mark.skip("No suitable dataset")
    def test_1s_Nf_u_16_12(self):
        pass

    # Non-conformant datasets
    def test_3s_1f_u_08_08_non(self):
        """Test process 2 greyscale."""
        # Non-conformant to DICOM
        ds = self.ds["JPEGExtended_3s_1f_u_08_08.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "YBR_FULL" in ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated
        assert 8 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        ds.pixel_array_options(raw=True)
        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        # Reference values from GDCM handler - in YBR colour space
        assert [
            [71, 86, 147],
            [124, 62, 167],
            [138, 66, 175],
            [145, 65, 172],
            [150, 59, 168],
            [178, 59, 167],
            [183, 56, 171],
            [201, 83, 165],
            [213, 92, 153],
            [255, 132, 134],
        ] == arr[41, 105:115].tolist()


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGLossless(HandlerTestBase):
    """Test the handler with ISO 10918 JPEG images.

    1.2.840.10008.1.2.4.57 : JPEG Lossless, Non-Hierarchical (Process 14)

    Process 14: Lossless, Huffman coding, 2 to 16-bit sample precision

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2, PALETTE COLOR,
        YBR_FULL, RGB
    Samples Per Pixel: 1 or 3
    Planar Configuration: 0 (for RGB/YBR_FULL)
    Pixel Representation: 0 (unsigned) or 1 (signed, monochrome only)
    Bits Allocated: 8 or 16
    Bits Stored: 1 to 16

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.1-2
      <part05/sect_8.2.html#table_8.2.1-2>`
    * ISO/IEC 10918-1 (ITU T.81)
    * ISO/IEC 10918-2
    * `JPEG Image Compression Systems
      <https://www.ece.ucdavis.edu/cerl/reliablejpeg/compression/>`_

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, signed, 1 frame
    1 sample/px, 8/8 bits allocated/stored, signed, > 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.57"

    def test_1s_1f_u_08_08(self):
        """Test process 2 greyscale."""
        ds = self.ds["JPEGLossless_1s_1f_u_16_12.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 12 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler - lossless must be identical!
        assert [392, 304, 238, 250, 224, 257, 221, 182, 166, 68] == (
            arr[779, 170:180].tolist()
        )


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGLosslessSV1(HandlerTestBase):
    """Test the handler with ISO 10918 JPEG images.

    1.2.840.10008.1.2.4.70 : JPEG Lossless, Non-Hierarchical, First-Order
    Prediction (Process 14 [Selection Value 1]

    Process 14: Lossless, Huffman coding, 2 to 16-bit sample precision,
    selection value of 1

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2, PALETTE COLOR,
        YBR_FULL, RGB
    Samples Per Pixel: 1 or 3
    Planar Configuration: 0 (for RGB/YBR_FULL)
    Pixel Representation: 0 (unsigned) or 1 (signed, monochrome only)
    Bits Allocated: 8 or 16
    Bits Stored: 1 to 16

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.1-2
      <part05/sect_8.2.html#table_8.2.1-2>`
    * ISO/IEC 10918-1 (ITU T.81)
    * ISO/IEC 10918-2
    * `JPEG Image Compression Systems
      <https://www.ece.ucdavis.edu/cerl/reliablejpeg/compression/>`_

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, signed, 1 frame
    1 sample/px, 8/8 bits allocated/stored, signed, > 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.70"

    # 1 sample/px, 8/8
    def test_1s_1f_u_08_08(self):
        """Test process 2 greyscale."""
        ds = self.ds["JPEGLosslessP14SV1_1s_1f_u_08_08.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated
        assert 8 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [0, 255, 254, 253, 253, 252, 251] == arr[121:128, 974].tolist()

    # 1 sample/px, 16/NN
    def test_1s_1f_u_16_10(self):
        """Test process 2 greyscale."""
        ds = self.ds["RG2_JPLL.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 10 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [574, 669, 763, 782, 789, 736, 637, 583, 592, 589] == (
            arr[845, 965:975].tolist()
        )

    def test_1s_1f_u_16_12(self):
        """Test process 2 greyscale."""
        ds = self.ds["MG1_JPLL.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 12 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [3023, 2862, 2841, 2825, 2841] == arr[1830:1835, 1285].tolist()

    def test_1s_1f_u_16_15(self):
        """Test process 2 greyscale."""
        ds = self.ds["RG1_JPLL.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 15 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [24043, 23906, 23669, 23526, 22962, 22874, 22501, 22066] == (
            arr[1900, 1805:1813].tolist()
        )

    def test_1s_1f_i_16_16(self):
        """Test process 2 greyscale."""
        ds = self.ds["JPEG-LL.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 16 == ds.BitsStored
        assert 1 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "int16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        assert 227 == arr[420, 140]
        assert 105 == arr[230, 120]

    # 3 samples/px, 8/8
    def test_3s_1f_u_08_08_rgb(self):
        """Test process 2 greyscale."""
        ds = self.ds["SC_rgb_jpeg_gdcm.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 3 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "RGB" in ds.PhotometricInterpretation
        assert 8 == ds.BitsAllocated
        assert 8 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint8" == arr.dtype
        assert (ds.Rows, ds.Columns, 3) == arr.shape

        # self.plot(arr)

        assert (255, 0, 0) == tuple(arr[5, 50, :])
        assert (255, 128, 128) == tuple(arr[15, 50, :])
        assert (0, 255, 0) == tuple(arr[25, 50, :])
        assert (128, 255, 128) == tuple(arr[35, 50, :])
        assert (0, 0, 255) == tuple(arr[45, 50, :])
        assert (128, 128, 255) == tuple(arr[55, 50, :])
        assert (0, 0, 0) == tuple(arr[65, 50, :])
        assert (64, 64, 64) == tuple(arr[75, 50, :])
        assert (192, 192, 192) == tuple(arr[85, 50, :])
        assert (255, 255, 255) == tuple(arr[95, 50, :])


# ISO/IEC 14495 JPEG-LS
@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGLSLossless(HandlerTestBase):
    """Test the handler with ISO 14495 JPEG-LS images.

    1.2.840.10008.1.2.4.80 : JPEG-LS Lossless Image Compression

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2, PALETTE COLOR,
        YBR_FULL, RGB
    Samples Per Pixel: 1 or 3
    Planar Configuration: 1 (for RGB/YBR_FULL)
    Pixel Representation: 0 (unsigned) or 1 (signed, monochrome only)
    Bits Allocated: 8 or 16
    Bits Stored: 2 to 16

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.3-1
      <part05/sect_8.2.html#table_8.2.3-1>`
    * ISO/IEC 14495

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, signed, 1 frame
    1 sample/px, 8/8 bits allocated/stored, signed, > 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, unsigned, 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, PALETTE COLOR, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.80"

    def test_1s_1f_i_16_16(self):
        """Test process 2 greyscale."""
        ds = self.ds["MR_small_jpeg_ls_lossless.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 16 == ds.BitsStored
        assert 1 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "int16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr, cmap='gray')

        # Reference values from GDCM handler
        assert [1194, 879, 127, 661, 1943, 1885, 1857, 1746, 1699] == (
            arr[55:65, 38].tolist()
        )

    def test_1s_Nf_u_16_12(self):
        """Test process 2 greyscale."""
        ds = self.ds["emri_small_jpeg_ls_lossless.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 10 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 12 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (10, ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr, index=0, cmap='gray')

        # Reference values from GDCM handler
        assert [361, 295, 215, 98, 79, 70, 41, 29, 61, 64] == (
            arr[0, 18:28, 48].tolist()
        )
        assert [298, 332, 355, 361, 324, 263, 169, 105, 53, 45] == (
            arr[9, 18:28, 48].tolist()
        )


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEGLS(HandlerTestBase):
    """Test the handler with ISO 14495 JPEG-LS images.

    Transfer Syntaxes
    -----------------

    1.2.840.10008.1.2.4.81 : JPEG-LS Lossy (Near-Lossless) Image Compression

    Supported Values
    ----------------
    Photometric Interpretation: MONOCHROME1, MONOCHROME2, YBR_FULL, RGB
    Samples Per Pixel: 1 or 3
    Planar Configuration: 1 (for RGB/YBR_FULL)
    Pixel Representation: 0 (unsigned) or 1 (signed, monochrome only)
    Bits Allocated: 8 or 16
    Bits Stored: 2 to 16

    References
    ----------
    * :dcm:`DICOM Standard, Part 5, Table 8.2.3-1
      <part05/sect_8.2.html#table_8.2.3-1>`
    * ISO/IEC 14495

    Missing datasets
    ----------------
    1 sample/px, 8/8 bits allocated/stored, signed, 1 frame
    1 sample/px, 8/8 bits allocated/stored, signed, > 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 8/8 bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, 1 frame
    1 sample/px, 16/xx bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, signed, > 1 frame
    1 sample/px, 16/xx bits allocated/stored, unsigned, > 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, 1 frame
    1 sample/px, 16/16 bits allocated/stored, unsigned, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, YBR_FULL, > 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 8/8 bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/xx bits allocated/stored, unsigned, RGB, > 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, 1 frame
    3 sample/px, 16/16 bits allocated/stored, unsigned, RGB, > 1 frame
    """

    uid = "1.2.840.10008.1.2.4.81"

    def test_1s_1f_i_16_16(self):
        """Test process 2 greyscale."""
        ds = self.ds["CT1_JLSN.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 16 == ds.BitsStored
        assert 1 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "int16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [-2000, -2000, 936, 946, 927, 925, 918, 929, 931, 921] == (
            arr[433, 69:79].tolist()
        )

    def test_1s_1f_u_16_10(self):
        """Test process 2 greyscale."""
        ds = self.ds["RG2_JLSN.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 10 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [574, 666, 760, 781, 789, 733, 634, 585, 594, 589] == (
            arr[845, 965:975].tolist()
        )

    def test_1s_1f_u_16_12(self):
        """Test process 2 greyscale."""
        ds = self.ds["MG1_JLSN.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 12 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [3027, 2859, 2839, 2821, 2837, 2834, 2817, 2844, 2837] == (
            arr[1830:1839, 1285].tolist()
        )

    def test_1s_1f_u_16_15(self):
        """Test process 2 greyscale."""
        ds = self.ds["RG1_JLSN.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 15 == ds.BitsStored
        assert 0 == ds.PixelRepresentation

        arr = ds.pixel_array
        assert arr.flags.writeable
        assert "uint16" == arr.dtype
        assert (ds.Rows, ds.Columns) == arr.shape

        # self.plot(arr)

        # Reference values from GDCM handler
        assert [24047, 23910, 23668, 23529, 22960, 22871, 22505, 22066] == (
            arr[1900, 1805:1813].tolist()
        )


# ISO/IEC 15444 JPEG 2000 - Expected fail
@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEG2000Lossless(HandlerTestBase):
    """Test the handler with ISO 15444 JPEG2000 images.

    1.2.840.10008.1.2.4.90 : JPEG 2000 Image Compression (Lossless Only)
    """

    uid = "1.2.840.10008.1.2.4.90"

    def test_1s_1f_i_16_16(self):
        """Test process 2 greyscale."""
        ds = self.ds["693_J2KR.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 16 == ds.BitsStored
        assert 1 == ds.PixelRepresentation

        msg = (
            r"Unable to decompress 'JPEG 2000 Image Compression \(Lossless Only\)' "
            "pixel data because all plugins are missing dependencies"
        )
        with pytest.raises(RuntimeError, match=msg):
            ds.pixel_array


@pytest.mark.skipif(not HAS_PYDICOM, reason="No dependencies")
class TestJPEG2000(HandlerTestBase):
    """Test the handler with ISO 15444 JPEG2000 images.

    1.2.840.10008.1.2.4.91 : JPEG 2000 Image Compression
    """

    uid = "1.2.840.10008.1.2.4.91"

    def test_1s_1f_i_16_16(self):
        """Test process 2 greyscale."""
        ds = self.ds["693_J2KI.dcm"]["ds"]
        assert self.uid == ds.file_meta.TransferSyntaxUID
        assert 1 == ds.SamplesPerPixel
        assert 1 == getattr(ds, "NumberOfFrames", 1)
        assert "MONOCHROME" in ds.PhotometricInterpretation
        assert 16 == ds.BitsAllocated
        assert 14 == ds.BitsStored
        assert 1 == ds.PixelRepresentation

        msg = (
            r"Unable to decompress 'JPEG 2000 Image Compression' "
            "pixel data because all plugins are missing dependencies"
        )
        with pytest.raises(RuntimeError, match=msg):
            ds.pixel_array



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['libjpeg', 'libjpeg.tests']

package_data = \
{'': ['*']}

install_requires = \
['numpy']

entry_points = \
{'pylibjpeg.jpeg_decoders': ['libjpeg = libjpeg:decode'],
 'pylibjpeg.jpeg_ls_decoders': ['libjpeg = libjpeg:decode'],
 'pylibjpeg.jpeg_xt_decoders': ['libjpeg = libjpeg:decode'],
 'pylibjpeg.pixel_data_decoders': ['1.2.840.10008.1.2.4.50 = '
                                   'libjpeg:decode_pixel_data',
                                   '1.2.840.10008.1.2.4.51 = '
                                   'libjpeg:decode_pixel_data',
                                   '1.2.840.10008.1.2.4.57 = '
                                   'libjpeg:decode_pixel_data',
                                   '1.2.840.10008.1.2.4.70 = '
                                   'libjpeg:decode_pixel_data',
                                   '1.2.840.10008.1.2.4.80 = '
                                   'libjpeg:decode_pixel_data',
                                   '1.2.840.10008.1.2.4.81 = '
                                   'libjpeg:decode_pixel_data']}

setup_kwargs = {
    'name': 'pylibjpeg-libjpeg',
    'version': '2.1.1',
    'description': 'A Python wrapper for libjpeg, with a focus on use as a plugin for for pylibjpeg',
    'long_description': '<p align="center">\n<a href="https://github.com/pydicom/pylibjpeg-libjpeg/actions?query=workflow%3Aunit-tests"><img alt="Build status" src="https://github.com/pydicom/pylibjpeg-libjpeg/workflows/unit-tests/badge.svg"></a>\n<a href="https://codecov.io/gh/pydicom/pylibjpeg-libjpeg"><img alt="Test coverage" src="https://codecov.io/gh/pydicom/pylibjpeg-libjpeg/branch/main/graph/badge.svg"></a>\n<a href="https://pypi.org/project/pylibjpeg-libjpeg/"><img alt="PyPI versions" src="https://img.shields.io/pypi/v/pylibjpeg-libjpeg"></a>\n<a href="https://www.python.org/"><img alt="Python versions" src="https://img.shields.io/pypi/pyversions/pylibjpeg-libjpeg"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\n## pylibjpeg-libjpeg\n\nA Python 3.8+ wrapper for Thomas Richter\'s\n[libjpeg](https://github.com/thorfdbg/libjpeg), with a focus on use as a\nplugin for [pylibjpeg](http://github.com/pydicom/pylibjpeg).\n\nLinux, MacOS and Windows are all supported.\n\n### Installation\n#### Dependencies\n[NumPy](http://numpy.org)\n\n#### Installing the current release\n```bash\npip install pylibjpeg-libjpeg\n```\n#### Installing the development version\n\nMake sure [Python](https://www.python.org/) and [Git](https://git-scm.com/) are installed. For Windows, you also need to install\n[Microsoft\'s C++ Build Tools](https://visualstudio.microsoft.com/thank-you-downloading-visual-studio/?sku=BuildTools&rel=16).\n```bash\ngit clone --recurse-submodules https://github.com/pydicom/pylibjpeg-libjpeg\npython -m pip install pylibjpeg-libjpeg\n```\n\n### Supported JPEG Formats\n#### Decoding\n\n| ISO/IEC Standard | ITU Equivalent | JPEG Format |\n| --- | --- | --- |\n| [10918](https://www.iso.org/standard/18902.html) | [T.81](https://www.itu.int/rec/T-REC-T.81/en) | [JPEG](https://jpeg.org/jpeg/index.html)    |\n| [14495](https://www.iso.org/standard/22397.html)   | [T.87](https://www.itu.int/rec/T-REC-T.87/en) | [JPEG-LS](https://jpeg.org/jpegls/index.html) |\n| [18477](https://www.iso.org/standard/62552.html)   | | [JPEG XT](https://jpeg.org/jpegxt/) |\n\n#### Encoding\nEncoding of JPEG images is not currently supported\n\n### Supported Transfer Syntaxes\n#### Decoding\n| UID | Description |\n| --- | --- |\n| 1.2.840.10008.1.2.4.50 | JPEG Baseline (Process 1) |\n| 1.2.840.10008.1.2.4.51 | JPEG Extended (Process 2 and 4) |\n| 1.2.840.10008.1.2.4.57 | JPEG Lossless, Non-Hierarchical (Process 14) |\n| 1.2.840.10008.1.2.4.70 | JPEG Lossless, Non-Hierarchical, First-Order Prediction (Process 14 [Selection Value 1]) |\n| 1.2.840.10008.1.2.4.80 | JPEG-LS Lossless |\n| 1.2.840.10008.1.2.4.81 | JPEG-LS Lossy (Near-Lossless) Image Compression |\n\n### Usage\n#### With pylibjpeg and pydicom\n\n```python\nfrom pydicom import dcmread\nfrom pydicom.data import get_testdata_file\n\nds = dcmread(get_testdata_file(\'JPEG-LL.dcm\'))\narr = ds.pixel_array\n```\n\n#### Standalone JPEG decoding\n\nYou can also decode JPEG images to a [numpy ndarray][1]:\n\n[1]: https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html\n\n```python\nfrom libjpeg import decode\n\nwith open(\'filename.jpg\', \'rb\') as f:\n    # Returns a numpy array\n    arr = decode(f.read())\n\n# Or simply...\narr = decode(\'filename.jpg\')\n```\n',
    'author': 'pylibjpeg-libjpeg contributors',
    'author_email': 'None',
    'maintainer': 'scaramallion',
    'maintainer_email': 'scaramallion@users.noreply.github.com',
    'url': 'https://github.com/pydicom/pylibjpeg-libjpeg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
